# coding: utf-8
# create by tongshiwei on 2019-8-30
import re

import torch


def collect_params(_net, select):
    if select is None:
        return _net.parameters()
    pattern = re.compile(select)
    ret = [value for name, value in _net.named_parameters() if pattern.match(name)]
    return ret


def get_trainer(_net, optimizer, optimizer_params=None, select=None):
    assert isinstance(optimizer, str) or issubclass(optimizer, torch.optim.Optimizer)
    parameters = collect_params(_net, select)
    assert optimizer_params is None or isinstance(optimizer_params, dict)
    optimizer_params = {} if optimizer_params is None else optimizer_params

    if isinstance(optimizer, str):
        if optimizer == "sgd":
            optimizer_class = torch.optim.SGD
        elif optimizer == "adam":
            optimizer_class = torch.optim.Adam
        elif optimizer == "adagrad":
            optimizer_class = torch.optim.Adagrad
        elif optimizer == "rmsprop":
            optimizer_class = torch.optim.RMSprop
        else:
            optimizer_class = eval("torch.optim.%s" % optimizer)
    else:
        optimizer_class = optimizer

    return optimizer_class(parameters, **optimizer_params)



# coding: utf-8
# create by tongshiwei on 2020-11-17

from copy import deepcopy

__all__ = ["get_optimizer_cfg"]

# By default, we suppose the batch_size is 16
# if the batch_size is 128, learning_rate should be 10 times default value
# e.g., for Adam, learning rate should be changed from 1e-3 to 1e-2.

optimizers = {
    "base": (
        "Adam",
        {
            "learning_rate": 1e-3,
            "wd": 1e-4,
            "clip_gradient": 1,
        }
    ),
    "recurrent": (
        "RMSProp",
        {
            "learning_rate": 1e-4,
            "wd": 1e-4,
            "clip_gradient": 1,
        }
    ),
    "stable": (
        "sgd",
        {
            "learning_rate": 1e-2,
            "wd": 1e-4,
            "clip_gradient": 1,
            "momentum": 0.9,
        }
    ),
    "word_embedding": (
        "adagrad",
        {
            "learning_rate": 1e-2,
            "wd": 1e-4,
            "clip_gradient": 1,
        }
    )
}


def get_optimizer_cfg(name):
    try:
        optimizer, optimizer_params = optimizers[name]
    except KeyError:
        raise KeyError("the name should be in: %s" % ", ".join(optimizers))
    optimizer_params = deepcopy(optimizer_params)

    return optimizer, optimizer_params



# coding: utf-8
# 2020/8/18 @ tongshiwei

import logging
import numpy as np
import scipy.stats as st
from mxnet.lr_scheduler import LRScheduler, FactorScheduler, PolyScheduler, MultiFactorScheduler, CosineScheduler

__all__ = ["get_lr_scheduler", "plot_schedule"]


def get_total_update_steps(update_epoch, batches_per_epoch, epoch_update_freq=1):
    return int(update_epoch * batches_per_epoch * epoch_update_freq)


def _stop_lr(base_lr, stop_lr=None, discount=None):
    assert stop_lr or discount
    if stop_lr:
        return stop_lr, discount
    elif discount:
        return base_lr * discount, discount


def get_factor_lr_params(base_lr, update_epoch, batches_per_epoch, epoch_update_freq=1, stop_lr=None, discount=None):
    assert epoch_update_freq

    total_update_steps = get_total_update_steps(
        update_epoch,
        batches_per_epoch,
        epoch_update_freq
    )
    step = batches_per_epoch // epoch_update_freq
    stop_lr, discount = _stop_lr(base_lr, stop_lr, discount)
    return step, pow(discount, step / total_update_steps), stop_lr


class _LRScheduler(LRScheduler):
    def __repr__(self):
        return str(self.state_dict)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def state_dict(self):
        raise NotImplementedError

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):
        if warmup_epoch:
            kwargs.update({"warmup_steps": warmup_epoch * batches_per_epoch})

        if discount is not None:
            final_lr, discount = _stop_lr(base_lr, discount=discount)
            kwargs.update({"final_lr": final_lr})

        return kwargs


class _FactorScheduler(FactorScheduler, _LRScheduler):
    def __repr__(self):
        return str(self.state_dict)

    @property
    def state_dict(self):
        return {
            "scheduler": "FactorScheduler",
            "base_lr": self.base_lr,
            "step": self.step,
            "stop_factor_lr": self.stop_factor_lr,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, **kwargs):
        kwargs = super(_FactorScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        )

        if batches_per_epoch is not None and update_epoch is not None:
            step, factor, stop_factor_lr = get_factor_lr_params(
                base_lr,
                update_epoch,
                batches_per_epoch,
                epoch_update_freq,
                kwargs.get("final_lr", kwargs.get("stop_factor_lr", None)),
                discount,
            )
            kwargs.update({"step": step, "factor": factor, "stop_factor_lr": stop_factor_lr})

        return cls(base_lr=base_lr, **kwargs)


class _PolyScheduler(PolyScheduler, _LRScheduler):
    @property
    def state_dict(self):
        return {
            "scheduler": "PolyScheduler",
            "base_lr": self.base_lr,
            "step": 1,
            "max_steps": self.max_steps,
            "final_lr": self.final_lr,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):
        kwargs = super(_PolyScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        )

        if batches_per_epoch is not None and update_epoch is not None:
            max_update = batches_per_epoch * update_epoch + kwargs.get("warmup_steps", 0)
            kwargs.update({"max_update": max_update})

        return cls(base_lr=base_lr, **kwargs)


class _CosineScheduler(CosineScheduler, _LRScheduler):
    @property
    def state_dict(self):
        return {
            "scheduler": "CosineScheduler",
            "base_lr": self.base_lr,
            "step": 1,
            "max_steps": self.max_steps,
            "final_lr": self.final_lr,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):
        kwargs.update(super(_CosineScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        ))

        if batches_per_epoch is not None and update_epoch is not None:
            max_update = batches_per_epoch * update_epoch * epoch_update_freq + kwargs.get("warmup_steps", 0)
            kwargs.update({"max_update": max_update})

        return cls(base_lr=base_lr, **kwargs)


class _MultiFactorScheduler(MultiFactorScheduler, _LRScheduler):
    @property
    def state_dict(self):
        return {
            "scheduler": "MultiFactorScheduler",
            "base_lr": self.base_lr,
            "max_steps": self.step,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):
        kwargs = super(_MultiFactorScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        )
        return cls(base_lr=base_lr, **kwargs)


class LinearScheduler(LRScheduler):
    def __init__(self, max_update, base_lr=0.01, final_lr=0,
                 warmup_steps=0, warmup_begin_lr=0, warmup_mode='linear'):
        super(LinearScheduler, self).__init__(
            base_lr, warmup_steps=warmup_steps, warmup_begin_lr=warmup_begin_lr, warmup_mode=warmup_mode
        )
        assert isinstance(max_update, int)
        if max_update < 1:
            raise ValueError("maximum number of updates must be strictly positive")
        self.max_update = max_update
        self.final_lr = final_lr
        self.max_steps = self.max_update - self.warmup_steps
        self.increase = (self.final_lr - self.base_lr) / float(self.max_steps)

    def __call__(self, num_update):
        if num_update < self.warmup_steps:
            return self.get_warmup_lr(num_update)
        self.base_lr += self.increase
        if self.base_lr < self.final_lr:
            return self.final_lr
        return self.base_lr


class _LinearScheduler(LinearScheduler, _LRScheduler):
    @property
    def state_dict(self):
        return {
            "scheduler": "LinearScheduler",
            "base_lr": self.base_lr,
            "max_steps": self.max_steps,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):
        kwargs = super(_LinearScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        )
        if batches_per_epoch is not None and update_epoch is not None:
            max_update = batches_per_epoch * update_epoch * epoch_update_freq + kwargs.get("warmup_steps", 0)
            kwargs.update({"max_update": max_update})
        return cls(base_lr=base_lr, **kwargs)


class NormScheduler(LRScheduler):
    def __init__(self,
                 max_update, step=1,
                 base_lr=0.01, final_lr=1e-8,
                 lr_loc=None, lr_scale=None,
                 min_lr=None, max_lr=None,
                 epsilon=1e-10,
                 warmup_steps=0, warmup_begin_lr=0, warmup_mode='linear'):

        if lr_loc is not None:
            self.lr_loc = lr_loc
            self.lr_scale = lr_scale if lr_scale is not None else 2 * lr_loc
            base_lr = self.lr_loc + 2 * self.lr_scale
        else:
            self.lr_loc = base_lr / 2 if lr_loc is None else lr_loc
            self.lr_scale = lr_scale if lr_scale is not None else (base_lr - self.lr_loc) / 2
        self.min_lr = max(final_lr, min_lr if min_lr is not None else (-2 * self.lr_scale + self.lr_loc))
        self.max_lr = min(max_lr if max_lr is not None else (2 * self.lr_scale + self.lr_loc), base_lr)
        self.dis = st.norm(loc=self.lr_loc, scale=self.lr_scale)
        self._p = 0
        self.epsilon = epsilon
        super(NormScheduler, self).__init__(
            base_lr=self.get_lr(), warmup_steps=warmup_steps,
            warmup_begin_lr=warmup_begin_lr, warmup_mode=warmup_mode,
        )
        self.step = step
        self.p_increase = self.step / (max_update - warmup_steps)
        self.count = 0

    @property
    def p(self):
        return np.clip(self.epsilon, self._p, 1 - self.epsilon)

    def get_lr(self):
        lr = self.dis.isf(self.p)
        if not self.min_lr <= lr <= self.max_lr:
            lr = np.clip(lr, self.min_lr, self.max_lr)
        return lr

    def __call__(self, num_update):
        if num_update < self.warmup_steps:
            return self.get_warmup_lr(num_update)

        # NOTE: use while rather than if  (for continuing training via load_epoch)
        while num_update - self.warmup_steps > self.count + self.step:
            self.count += self.step
            self._p += self.p_increase
            self.base_lr = self.get_lr()
            logging.info("Update[%d]: Change learning rate to %0.5e",
                         num_update, self.base_lr)
        return self.base_lr


class _NormScheduler(NormScheduler, _LRScheduler):
    @property
    def state_dict(self):
        return {
            "scheduler": "NormScheduler",
            "lr_loc": self.lr_loc,
            "step": self.step,
            "max_lr": self.max_lr,
            "min_lr": self.min_lr,
            "lr_scale": self.lr_scale,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr=None, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):
        kwargs.update(super(_NormScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount,
        ))
        if batches_per_epoch is not None and update_epoch is not None:
            max_update = int(batches_per_epoch * update_epoch) + kwargs.get("warmup_steps", 0)
            kwargs.update({"max_update": max_update, "step": batches_per_epoch // epoch_update_freq})
        return cls(base_lr=base_lr, **kwargs)


SCHEDULERS = {
    "linear": _LinearScheduler,
    "factor": _FactorScheduler,
    "poly": _PolyScheduler,
    "multifactor": _MultiFactorScheduler,
    "cosine": _CosineScheduler,
    "norm": _NormScheduler,
}


def get_lr_scheduler(scheduler: (str, LRScheduler) = "cosine", logger=logging, update_params=None, **kwargs):
    for key in {"learning_rate", "lr"}:
        if key in kwargs:
            kwargs["base_lr"] = kwargs.pop(key)

    if isinstance(scheduler, str):
        scheduler = SCHEDULERS[scheduler].init(**kwargs)

    assert isinstance(scheduler, LRScheduler)

    logger.info(scheduler)

    return scheduler


def plot_schedule(schedule_fn, iterations=1500):
    import matplotlib.pyplot as plt
    # Iteration count starting at 1
    iterations = [i + 1 for i in range(iterations)]
    lrs = [schedule_fn(i) for i in iterations]
    plt.scatter(iterations, lrs)
    plt.xlabel("Iteration")
    plt.ylabel("Learning Rate")
    plt.show()


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.INFO)
    _scheduler = get_lr_scheduler(
        "norm", base_lr=1, discount=0.01, lr_loc=0.02,
        batches_per_epoch=100, update_epoch=10, epoch_update_freq=10,
    )
    plot_schedule(_scheduler)



# coding: utf-8
# create by tongshiwei on 2018/7/14

from .ValueMonitor import ValueMonitor, EMAValue, tmt_value, as_tmt_value

__all__ = ["LossMonitor", "MovingLoss", "tmt_loss", "as_tmt_loss", "loss_dict2tmt_loss"]


def tmt_loss(loss2value=lambda x: x):
    return tmt_value(transform=loss2value)


def as_tmt_loss(loss_obj, loss2value=lambda x: x):
    return as_tmt_value(loss_obj, loss2value)


def loss_dict2tmt_loss(loss_dict, loss2value=lambda x: x, exclude=None, include=None, as_loss=as_tmt_loss):
    exclude = set() if exclude is None else set(exclude)
    if include is not None:
        include = set(include)
        return {
            name: as_loss(func, loss2value) if name in include else func for name, func in loss_dict.items()
        }
    return {
        name: as_loss(func, loss2value) if name not in exclude else func for name, func in loss_dict.items()
    }


class LossMonitor(ValueMonitor):
    @property
    def losses(self):
        return self.value


class MovingLoss(EMAValue, LossMonitor):
    """
    Examples
    --------
    >>> lm = MovingLoss(["l2"])
    >>> lm.losses
    {'l2': nan}
    >>> lm("l2", 100)
    >>> lm("l2", 1)
    >>> lm["l2"]
    90.1
    """
    pass



# coding: utf-8
# create by tongshiwei on 2019/4/12
import warnings

import mxnet as mx

from longling.ML.DL import get_optimizer_cfg

__all__ = ["get_optimizer_cfg", "get_lr_scheduler", "get_update_steps"]


def get_lr_params(batches_per_epoch, lr, update_epoch, epoch_update_freq=1, *args, **kwargs):
    warnings.warn(
        "deprecated since version 1.3.13, switched to longling.ML.toolkit.lr_scheduler.get_lr_scheduler"
    )
    return {
        "learning_rate": lr,
        "step": batches_per_epoch // epoch_update_freq,
        "max_update_steps": get_update_steps(
            update_epoch=update_epoch,
            batches_per_epoch=batches_per_epoch,
            epoch_update_freq=epoch_update_freq,
        ),
    }


def get_update_steps(update_epoch, batches_per_epoch, epoch_update_freq=1):
    warnings.warn(
        "deprecated since version 1.3.13, switched to longling.ML.toolkit.lr_scheduler.get_lr_scheduler"
    )
    return int(update_epoch * batches_per_epoch * epoch_update_freq)


def get_lr_scheduler(learning_rate, step=None, max_update_steps=None,
                     discount=0.01):
    warnings.warn(
        "deprecated since version 1.3.13, switched to longling.ML.toolkit.lr_scheduler.get_lr_scheduler"
    )
    factor = pow(discount, step / max_update_steps)
    return mx.lr_scheduler.FactorScheduler(
        step, factor,
        stop_factor_lr=learning_rate * discount
    )


def plot_schedule(schedule_fn, iterations=1500):
    import matplotlib.pyplot as plt
    # Iteration count starting at 1
    iterations = [i + 1 for i in range(iterations)]
    lrs = [schedule_fn(i) for i in iterations]
    plt.scatter(iterations, lrs)
    plt.xlabel("Iteration")
    plt.ylabel("Learning Rate")
    plt.show()



# coding: utf-8
# create by tongshiwei on 2020-11-21

from collections import Iterable

import pandas as pd
from longling import as_list
from numpy.random import RandomState
from tqdm import tqdm

"""
rating matrix
(query, key, value): value could be omit

contrastive triplet
(query, pos, neg)
"""

__all__ = ["TripletPairSampler", "UITripletPairSampler"]


class Sampler(object):
    def __init__(self, random_state=10, pad_value=0):
        self.random_state = RandomState(random_state)
        self.pad_value = pad_value

    def padding(self, sampled: list, n, *args, **kwargs) -> list:
        pad_value = kwargs.get("pad_value", self.pad_value)
        return sampled + [pad_value] * (n - len(sampled))


class PairSampler(object):
    def __init__(self):
        pass


class TripletPairSampler(Sampler):
    def __init__(self, triplet_df: pd.DataFrame,
                 query_field, pos_field="pos", neg_field="neg", set_index=False,
                 query_range=None, key_range=None,
                 random_state=10):
        super(TripletPairSampler, self).__init__(random_state)
        self.df = triplet_df
        self.query_field = query_field
        self.pos_field = pos_field
        self.neg_field = neg_field
        self.query_range = [0, query_range]
        self.key_range = [0, key_range]
        if set_index:
            self.df.set_index(self.query_field, inplace=True)

    def __call__(self, query: (int, str, list), n=1, excluded_key=None, neg=True, implicit=False, padding=True,
                 *args, verbose=True, return_column=False, split_sample_to_column=False, **kwargs):
        if isinstance(query, (int, str)):
            sample = self.sample(
                query=query,
                n=n,
                excluded_key=excluded_key,
                neg=neg
            ) if not implicit else self.implicit_sample(
                query=query,
                n=n,
                excluded_key=excluded_key,
                neg=neg
            )
            n_sample = len(sample)
            if padding:
                return n_sample, self.padding(sample, n, **kwargs)
            return n_sample, sample
        else:
            neg = neg if isinstance(neg, Iterable) else [neg] * len(query)
            excluded_key = excluded_key if isinstance(excluded_key, Iterable) else [excluded_key] * len(query)
            n_and_sample = [
                self(_query, n, _excluded_key, _neg, implicit, padding, *args, **kwargs)
                for _query, _excluded_key, _neg in tqdm(
                    zip(query, excluded_key, neg), "sampling", disable=not verbose
                )
            ]
            if return_column:
                n_sample, sample = zip(*n_and_sample)
                if split_sample_to_column:
                    return n_sample, list(zip(*sample))
                return n_sample, sample
            return n_and_sample

    def sample(self, query: (int, str, list), n=1, excluded_key=None, neg=True, *args, **kwargs):
        try:
            candidates = self.df.loc[query][self.neg_field] if neg else self.df.loc[query][self.pos_field]
        except IndexError as e:
            print(query, excluded_key, neg)
            raise e

        if excluded_key is not None:
            candidates = list(set(candidates) - set(as_list(excluded_key)))

        sampled = self.random_state.choice(candidates, min(n, len(candidates)), replace=False).tolist()
        return sampled

    def implicit_sample(self, query: (int, str, list), n=1, excluded_key=None, neg=True, *args, **kwargs):
        exclude = set(self.df.loc[query][self.pos_field] if neg else self.df.loc[query][self.neg_field])
        if excluded_key is not None:
            exclude |= set(as_list(excluded_key))
        candidates = list(set(range(*self.key_range)) - exclude)
        sampled = self.random_state.choice(candidates, min(n, len(candidates)), replace=False).tolist()
        return sampled

    @staticmethod
    def rating2triplet(rating_matrix: (pd.DataFrame, list), query_field, key_field, value_field=None,
                       value_threshold=None,
                       pos_field="pos", neg_field="neg",
                       *args, verbose=True, **kwargs):
        if isinstance(rating_matrix, pd.DataFrame):
            triplet = []
            for _, group in tqdm(rating_matrix.groupby(query_field), "rating2triplet", disable=not verbose):
                query_key = group[query_field].unique()[0]
                if value_field is None:
                    pos = group[key_field]
                    neg = pd.Series([])
                elif value_threshold:
                    pos = group[group[value_field] >= value_threshold][key_field]
                    neg = group[group[value_field] < value_threshold][key_field]
                else:
                    pos = group[group[value_field] == 1][key_field]
                    neg = group[group[value_field] == 0][key_field]
                triplet.append([query_key, pos.tolist(), neg.tolist()])
        elif isinstance(rating_matrix, list):
            triplet = rating_matrix
        else:
            raise NotImplementedError

        df = pd.DataFrame(triplet, columns=[query_field, pos_field, neg_field])
        return df.set_index(query_field)


class UITripletPairSampler(TripletPairSampler):
    """
    User-Item

    Examples
    --------
    >>> import pandas as pd
    >>> user_num = 3
    >>> item_num = 4
    >>> rating_matrix = pd.DataFrame({
    ...     "user_id": [0, 1, 1, 1, 2],
    ...     "item_id": [1, 3, 0, 2, 1]
    ... })
    >>> triplet_df = UITripletPairSampler.rating2triplet(rating_matrix)
    >>> triplet_df.index
    Int64Index([0, 1, 2], dtype='int64', name='user_id')
    >>> sampler = UITripletPairSampler(triplet_df)
    >>> sampler(1)
    (0, [0])
    >>> sampler = UITripletPairSampler(triplet_df, item_id_range=item_num)
    >>> sampler(0, implicit=True)
    (1, [0])
    >>> sampler(0, 5, implicit=True)
    (3, [3, 2, 0, 0, 0])
    >>> sampler(0, 5, implicit=True, pad_value=-1)
    (3, [3, 2, 0, -1, -1])
    >>> sampler([0, 1, 2], 5, implicit=True, pad_value=-1)
    [(3, [3, 2, 0, -1, -1]), (1, [1, -1, -1, -1, -1]), (3, [3, 0, 2, -1, -1])]
    >>> rating_matrix = pd.DataFrame({
    ...     "user_id": [0, 1, 1, 1, 2],
    ...     "item_id": [1, 3, 0, 2, 1],
    ...     "score": [1, 0, 1, 1, 0]
    ... })
    >>> triplet_df = UITripletPairSampler.rating2triplet(rating_matrix=rating_matrix, value_field="score")
    >>> triplet_df["pos"]
    user_id
    0       [1]
    1    [0, 2]
    2        []
    Name: pos, dtype: object
    >>> triplet_df["neg"]
    user_id
    0     []
    1    [3]
    2    [1]
    Name: neg, dtype: object
    >>> sampler = UITripletPairSampler(triplet_df)
    >>> sampler([0, 1, 2], 5, pad_value=-1)
    [(0, [-1, -1, -1, -1, -1]), (1, [3, -1, -1, -1, -1]), (1, [1, -1, -1, -1, -1])]
    >>> sampler([0, 1, 2], 5, neg=False, pad_value=-1)
    [(1, [1, -1, -1, -1, -1]), (2, [0, 2, -1, -1, -1]), (0, [-1, -1, -1, -1, -1])]
    >>> sampler(rating_matrix["user_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["item_id"], pad_value=-1)
    [(0, [-1, -1]), (2, [0, 2]), (1, [3, -1]), (1, [3, -1]), (0, [-1, -1])]
    >>> sampler(rating_matrix["user_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["item_id"], pad_value=-1, return_column=True)
    ((0, 2, 1, 1, 0), ([-1, -1], [2, 0], [3, -1], [3, -1], [-1, -1]))
    >>> sampler(rating_matrix["user_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["item_id"], pad_value=-1, return_column=True, split_sample_to_column=True)
    ((0, 2, 1, 1, 0), [(-1, 0, 3, 3, -1), (-1, 2, -1, -1, -1)])
    """

    def __init__(self, triplet_df: pd.DataFrame,
                 query_field="user_id", pos_field="pos", neg_field="neg", set_index=False,
                 user_id_range=None, item_id_range=None,
                 random_state=10):
        super(UITripletPairSampler, self).__init__(
            triplet_df=triplet_df,
            query_field=query_field,
            pos_field=pos_field,
            neg_field=neg_field,
            set_index=set_index,
            query_range=user_id_range,
            key_range=item_id_range,
            random_state=random_state)

    @staticmethod
    def rating2triplet(rating_matrix: (pd.DataFrame, list),
                       query_field="user_id", key_field="item_id", value_field=None,
                       value_threshold=None, pos_field="pos", neg_field="neg",
                       *args, verbose=True, **kwargs):
        return TripletPairSampler.rating2triplet(
            rating_matrix=rating_matrix,
            query_field=query_field,
            key_field=key_field,
            value_field=value_field,
            value_threshold=value_threshold,
            pos_field=pos_field,
            neg_field=neg_field,
            *args,
            verbose=verbose,
            **kwargs
        )



# coding: utf-8
# create by tongshiwei on 2019/4/11

"""
自定义的配置文件及对应的解析工具包。目的是为了更方便、快速地进行文件参数配置与解析。
"""
import argparse
import inspect
import json
import logging
import os
import re
import sys
from collections import Iterable
from copy import deepcopy

import toml
import yaml
from longling import wf_open
from longling.lib.path import path_append

__all__ = [
    "CLASS_EXCLUDE_NAMES", "get_class_var",
    "get_parsable_var", "load_configuration",
    "var2exp", "path_append",
    "Configuration", "ConfigurationParser",
    "Formatter"
]

CLASS_EXCLUDE_NAMES = set(vars(object).keys()) | {
    '__module__', '__main__', '__dict__', '__weakref__',
}


def get_class_var(class_obj, exclude_names: (set, None) = None, get_vars=None) -> dict:
    """
    Update in v1.3.18

    获取某个类的所有属性的变量名及其值

    Examples
    --------
    >>> class A(object):
    ...     att1 = 1
    ...     att2 = 2
    >>> get_class_var(A)
    {'att1': 1, 'att2': 2}
    >>> get_class_var(A, exclude_names={"att1"})
    {'att2': 2}
    >>> class B(object):
    ...     att3 = 3
    ...     att4 = 4
    ...     @staticmethod
    ...     def excluded_names():
    ...         return {"att4"}
    >>> get_class_var(B)
    {'att3': 3}

    Parameters
    ----------
    class_obj:
        类或类实例。需要注意两者的区别。
    exclude_names:
        需要排除在外的变量名。也可以通过在类定义 excluded_names 方法来指定要排除的变量名。
    get_vars

    Returns
    -------
    class_var:
        类内属性变量名及值

    """
    default_excluded_names = CLASS_EXCLUDE_NAMES if exclude_names is None \
        else CLASS_EXCLUDE_NAMES | exclude_names
    excluded_names = default_excluded_names if not hasattr(
        class_obj, "excluded_names"
    ) else class_obj.excluded_names() | default_excluded_names

    variables = {}
    if get_vars is not False:  # True or None
        variables.update({
            k: v for k, v in vars(class_obj).items() if
            not inspect.isroutine(v) and k not in excluded_names
        })
    if get_vars is not True:  # False or None
        for k in dir(class_obj):
            if hasattr(class_obj, k):
                v = getattr(class_obj, k)
                if not inspect.isroutine(v) and k not in excluded_names:
                    variables[k] = v
    return variables


def parse_params(params: dict, parse_functions: (dict, None) = None) -> dict:
    """
    参数后处理，根据parse_functions中定义的转换函数，对params中的变量值进行转换。
    例如，可用于将无法直接dump的变量转换成可以dump的变量。

    Examples
    --------
    >>> params = {"a": 1}
    >>> parse_functions = {"a": lambda x: x + x}
    >>> parse_params(params, parse_functions)
    {'a': 2}
    >>> params = {"a": 1}
    >>> parse_params(params)
    {'a': 1}

    """
    if parse_functions is not None:
        for name, func in parse_functions.items():
            if name in params:
                params[name] = func(params[name])
    return params


def get_parsable_var(class_obj, parse_exclude: set = None, dump_parse_functions=None, get_vars=True):
    """获取所有可以被解析的参数及其值，可以使用dump_parse_functions来对不可dump的值进行转换"""
    params = get_class_var(class_obj, exclude_names=parse_exclude, get_vars=get_vars)
    return parse_params(params, dump_parse_functions)


def load_configuration(fp, file_format="json", load_parse_function=None):
    """
    装载配置文件

    Updated in version 1.3.16

    Parameters
    ----------
    fp
    file_format
    load_parse_function

    """
    if file_format == "json":
        params = json.load(fp)
    elif file_format == "toml":
        params = toml.load(fp)
    elif file_format == "yaml":
        params = yaml.load(fp)
    else:
        raise TypeError("Unsupported file format: %s, only `json`, `toml` and `yaml` are supported" % file_format)
    return parse_params(params, load_parse_function)


def var2exp(var_str, env_wrap=lambda x: x):
    """
    将含有 $ 标识的变量转换为表达式

    Parameters
    ----------
    var_str
    env_wrap

    Examples
    --------
    >>> root = "dir"
    >>> dataset = "d1"
    >>> eval(var2exp("$root/data/$dataset"))
    'dir/data/d1'
    """
    var_str = str(var_str)

    pattern = re.compile(r"\$(\w+)")

    env_vars = pattern.findall(var_str)

    exp = """str(r"%s").format(%s)""" % (
        re.sub(r"\$\w+", "{}", var_str), ",".join(
            [env_wrap(env_var) for env_var in env_vars]
        )
    )
    return exp


def is_classmethod(method):
    """

    Parameters
    ----------
    method

    Returns
    -------

    Examples
    --------
    >>> class A:
    ...     def a(self):
    ...         pass
    ...     @staticmethod
    ...     def b():
    ...         pass
    ...     @classmethod
    ...     def c(cls):
    ...         pass
    >>> obj = A()
    >>> is_classmethod(obj.a)
    False
    >>> is_classmethod(obj.b)
    False
    >>> is_classmethod(obj.c)
    True
    >>> def fun():
    ...     pass
    >>> is_classmethod(fun)
    False
    """
    bound_to = getattr(method, '__self__', None)
    if not isinstance(bound_to, type):
        # must be bound to a class
        return False
    name = method.__name__
    for cls in bound_to.__mro__:
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False


class Configuration(object):
    """自定义的配置文件基类"""

    def __init__(self, logger=logging, **kwargs):
        self.logger = logger

        params = self.class_var

        params.update(**kwargs)
        for param, value in params.items():
            # all class variables will be contained in instance variables
            setattr(self, "%s" % param, value)

    @classmethod
    def vars(cls):
        # todo: check why when deepcopy is missed, two instance will have same id for class_var
        return deepcopy(get_class_var(cls))

    @property
    def class_var(self):
        """
        获取所有设定的参数

        Returns
        -------
        parameters: dict
            all variables used as parameters
        """
        return self.vars()

    @classmethod
    def pvars(cls):
        return get_parsable_var(
            cls,
            parse_exclude=None,
            dump_parse_functions=None,
        )

    @property
    def parsable_var(self):
        """
        获取可以进行命令行设定的参数

        Returns
        -------
        store_vars: dict
            可以进行命令行设定的参数
        """
        return get_parsable_var(
            self,
            parse_exclude=None,
            dump_parse_functions=None,
        )

    @classmethod
    def load_parse_function(cls):
        return None

    @classmethod
    def default_file_format(cls):
        return "json"

    @classmethod
    def load_cfg(cls, cfg_path, file_format=None, **kwargs):
        """从配置文件中装载配置参数"""
        file_format = file_format if file_format is not None else cls.default_file_format()

        with open(cfg_path) as f:
            params = load_configuration(
                f, file_format=file_format, load_parse_function=cls.load_parse_function()
            )
        params.update(kwargs)
        return params

    @classmethod
    def load(cls, cfg_path, file_format=None, **kwargs):
        """
        从配置文件中装载配置类

        Updated in version 1.3.16
        """
        file_format = file_format if file_format is not None else cls.default_file_format()
        return cls(**cls.load_cfg(cfg_path, file_format=file_format, **kwargs))

    def dump(self, cfg_path: str, override=True, file_format=None):
        """
        将配置参数写入文件

        Updated in version 1.3.16

        Parameters
        ----------
        cfg_path: str
        override: bool
        file_format: str
        """
        if os.path.isfile(cfg_path) and not override:
            self.logger.warning(
                "file %s existed, dump aborted" % os.path.abspath(cfg_path)
            )
            return
        self.logger.info(
            "writing configuration parameters to %s" % os.path.abspath(cfg_path)
        )
        file_format = file_format if file_format is not None else self.default_file_format()

        with wf_open(cfg_path) as wf:
            if file_format == "json":
                json.dump(self.parsable_var, wf, indent=2)
            elif file_format == "toml":
                toml.dump(self.parsable_var, wf)
            elif file_format == "yaml":
                yaml.dump(self.parsable_var, wf)
            else:
                raise TypeError(
                    "Unsupported file format: %s, only `json`, `toml` and `yaml` are supported" % file_format
                )

    def __str__(self):
        string = []
        for k, v in vars(self).items():
            string.append("%s: %s" % (k, v))
        return "\n".join(string)

    def __repr__(self):
        return str(self)

    def __getitem__(self, item):
        return getattr(self, item)

    def __contains__(self, item):
        return item in self.class_var

    def items(self):
        return {k: v for k, v in self.parsable_var.items()}

    @classmethod
    def run_time_variables(cls):
        return {'logger'}

    @classmethod
    def excluded_names(cls):
        """
        获取非参变量集

        Returns
        -------
        exclude names set: set
            所有非参变量
        """
        return CLASS_EXCLUDE_NAMES | {
            'class_var', 'parsable_var', 'items', 'load', 'dump', 'help_info'
        } | cls.run_time_variables()

    @classmethod
    def help_info(cls) -> dict:
        return {}


class InheritableConfiguration(Configuration):
    @classmethod
    def vars(cls):
        return get_class_var(cls, get_vars=False)

    @property
    def parsable_var(self):
        """
        获取可以进行命令行设定的参数

        Returns
        -------
        store_vars: dict
            可以进行命令行设定的参数
        """
        return get_parsable_var(
            self,
            parse_exclude=None,
            dump_parse_functions=None,
            get_vars=False
        )


def value_parse(value):
    r"""
    将含有关键字的字符串转换为关键字指定的类型
    支持的关键字类型为python的基础数据类型:
    int, float, dict, list, set, tuple, None

    Parameters
    ----------
    value: str
        字符串值

    Examples
    --------
    >>> value_parse("int(1.0)")
    1
    >>> value_parse("float(1.0)")
    1.0
    >>> value_parse("dict(a=1, b=2.0, c='d')")
    {'a': 1, 'b': 2.0, 'c': 'd'}
    >>> value_parse("list([1, 2, 3, 4])")
    [1, 2, 3, 4]
    >>> value_parse("set([1, 1, 2, 3])")
    {1, 2, 3}
    >>> value_parse("tuple([1, 2, 3])")
    (1, 2, 3)
    >>> value_parse("None") is None
    True
    """
    if re.findall(r"(int|float|dict|list|set|tuple|bool)\(.*\)|None", value):
        value = eval(value)
    return value


def parse_dict_string(string):
    """
    解析字典字符串

    Examples
    --------
    >>> parse_dict_string("a=1;b=int(1);c=list([1]);d=None")
    {'a': '1', 'b': 1, 'c': [1], 'd': None}
    >>> parse_dict_string("")
    """
    if not string:
        return None
    else:
        name_value_items = [
            item.strip().split("=", 1) for item in string.strip().rstrip(";").split(";")
        ]
        return {
            name_value[0]: value_parse(name_value[1])
            for name_value in name_value_items
        }


def args_zips(args=None, defaults=None):
    """组合参数及默认值"""
    args = [] if args is None else args
    defaults = [] if defaults is None else defaults
    assert len(args) >= len(defaults)
    zipped_args = []
    zipped_args.extend([
        {"name": arg} for arg in args[:len(args) - len(defaults)]
    ])
    zipped_args.extend([
        {"name": arg, "default": default} for arg, default in zip(
            args[len(args) - len(defaults):], defaults
        )
    ])
    return zipped_args


class ConfigurationParser(argparse.ArgumentParser):
    """
    Update in v1.3.18

    配置文件解析类，可用于构建cli工具。该类首先读入所有目标配置文件类class_obj的所有类属性，解析后生成命令行。
    普通属性参数使用 "--att_name att_value" 来读入。另外提供一个额外参数标记 ‘--kwargs’ 来读入可选参数。
    可选参数格式为 ::

        --kwargs key1=value1;key2=value2;...

    首先生成一个解析类 ::

        cli_parser = ConfigurationParser(Configuration)

    除了解析已有配置文件外，解析类还可以进一步添加函数来生成子解析器 ::

        cli_parser = ConfigurationParser($function)

    或者 ::

        cli_parser = ConfigurationParser([$function1, $function2])

    用以下三种解析方式中任意一种来解析参数：

    * 命令行模式 ::

        cli_parser()

    * 字符串传参模式 ::

        cli_parser('$parameter1 $parameters ...')

    * 列表传参模式 ::

        cli_parser(["--a", "int(1)", "--b", "int(2)"])

    Notes
    -----
    包含以下关键字的字符串会在解析过程中进行类型转换

    int, float, dict, list, set, tuple, None

    Parameters
    ----------
    class_type:
        类。注意是类，不是类实例。
    excluded_names:
        类中不进行解析的变量名集合
    commands:
        待解析的命令函数

    Examples
    --------
    >>> class TestC(Configuration):
    ...     a = 1
    ...     b = 2
    >>> def test_f1(k=1):
    ...     return k
    >>> def test_f2(h=1):
    ...      return h
    >>> def test_f3(m):
    ...      return m
    >>> parser = ConfigurationParser(TestC)
    >>> parser("--a 1 --b 2")
    {'a': '1', 'b': '2'}
    >>> ConfigurationParser.get_cli_cfg(TestC)
    {'a': 1, 'b': 2}
    >>> parser(["--a", "1", "--b", "int(1)"])
    {'a': '1', 'b': 1}
    >>> parser(["--a", "1", "--b", "int(1)", "--kwargs", "c=int(3);d=None"])
    {'a': '1', 'b': 1, 'c': 3, 'd': None}
    >>> parser.add_command(test_f1, test_f2, test_f3)
    >>> parser(["test_f1"])
    {'a': 1, 'b': 2, 'k': 1, 'subcommand': 'test_f1'}
    >>> parser(["test_f2"])
    {'a': 1, 'b': 2, 'h': 1, 'subcommand': 'test_f2'}
    >>> parser(["test_f3", "3"])
    {'a': 1, 'b': 2, 'm': '3', 'subcommand': 'test_f3'}
    >>> parser = ConfigurationParser(TestC, commands=[test_f1, test_f2])
    >>> parser(["test_f1"])
    {'a': 1, 'b': 2, 'k': 1, 'subcommand': 'test_f1'}
    >>> class TestCC:
    ...     c = {"_c": 1, "_d": 0.1}
    >>> parser = ConfigurationParser(TestCC)
    >>> parser("--c _c=int(3);_d=float(0.3)")
    {'c': {'_c': 3, '_d': 0.3}}
    >>> class TestCls:
    ...     def a(self, a=1):
    ...         return a
    ...     @staticmethod
    ...     def b(b=2):
    ...         return b
    ...     @classmethod
    ...     def c(cls, c=3):
    ...         return c
    >>> parser = ConfigurationParser(TestCls, commands=[TestCls.b, TestCls.c])
    >>> parser("b")
    {'b': 2, 'subcommand': 'b'}
    >>> parser("c")
    {'c': 3, 'subcommand': 'c'}
    """

    def __init__(self, class_type, excluded_names: (set, None) = None,
                 commands=None,
                 *args,
                 params_help=None, commands_help=None, override_help=False,
                 **kwargs):
        excluded_names = {
            'logger'
        } if excluded_names is None else excluded_names

        self.__proto_type = argparse.ArgumentParser(add_help=False)

        params = {k: v for k, v in get_class_var(class_type).items()}
        _params_help = {}
        _params_help.update(class_type.help_info() if hasattr(class_type, 'help_info') else {})
        _params_help.update({} if params_help is None else params_help)
        params_help = _params_help

        def get_help_info(default_help_info, param_name=None):
            if not params_help or param_name not in params_help:
                return default_help_info
            else:
                if override_help:
                    return "%s" % params_help[param_name]
                else:
                    return "%s, %s" % (params_help[param_name], default_help_info)

        for param, value in params.items():
            if param in excluded_names:  # pragma: no cover
                continue
            if isinstance(value, dict):
                format_tips = ", dict variables, " \
                              "use format: <key>=<value>(;<key>=<value>)"
                self.__proto_type.add_argument(
                    '--%s' % param,
                    help=get_help_info('set %s, default is %s%s' % (
                        param, value, format_tips
                    ), param),
                    default=value,
                    type=parse_dict_string,
                )
            else:
                self.__proto_type.add_argument(
                    '--%s' % param,
                    help=get_help_info('set %s, default is %s' % (param, value), param),
                    default=value,
                    type=value_parse,
                )
        self.__proto_type.add_argument(
            '--kwargs', required=False,
            help=r"add extra argument here, "
                 r"use format: <key>=<value>(;<key>=<value>)"
        )
        self.sub_command_parsers = None
        super(ConfigurationParser, self).__init__(parents=[self.__proto_type], *args, **kwargs)

        if commands is not None:
            assert isinstance(commands, Iterable)
            self.add_command(*commands, help_info=commands_help)

    def add_command(self, *commands, help_info: (list, dict, str, None) = None):
        """批量添加子命令解析器"""
        if isinstance(help_info, list):
            for command, _help_info in zip(commands, help_info):
                self._add_subcommand(self.func_spec(command), _help_info)
        else:
            for command in commands:
                self._add_subcommand(self.func_spec(command), help_info)

    @staticmethod
    def parse(arguments):
        """参数后解析"""
        arguments = vars(arguments)
        args_dict = dict()
        for k, v in arguments.items():
            if k == "kwargs" and v:
                kwargs = parse_dict_string(v)
                if kwargs:
                    args_dict.update(kwargs)
            elif k == "kwargs":
                pass
            else:
                args_dict[k] = v
        return args_dict

    @classmethod
    def get_cli_cfg(cls, params_class) -> dict:
        """获取默认配置参数"""
        params_parser = cls(params_class)
        kwargs = params_parser.parse_args([])
        kwargs = params_parser.parse(kwargs)
        return kwargs

    def __call__(self, args: (str, list, None) = None) -> dict:
        if isinstance(args, str):
            args = args.split(" ")
        kwargs = self.parse_args(args)
        kwargs = self.parse(kwargs)
        return kwargs

    def _add_subcommand(self, command_parameters, help_info=None):
        """添加子命令解析器"""

        def get_help_info(default_help_info, command=None):
            if isinstance(help_info, dict):
                return help_info.get(command, default_help_info)
            elif command is None:
                return default_help_info
            else:
                return command

        if self.sub_command_parsers is None:
            self.sub_command_parsers = self.add_subparsers(
                parser_class=argparse.ArgumentParser,
                help='help for sub-command'
            )
        subparsers = self.sub_command_parsers
        subcommand, parameters = command_parameters
        subparser = subparsers.add_parser(
            subcommand, help=get_help_info("%s help" % subcommand, subcommand), parents=[self.__proto_type],
        )
        subparser.set_defaults(subcommand=subcommand)
        for parameter in parameters:
            if "default" in parameter:
                subparser.add_argument(
                    "--%s" % parameter["name"],
                    default=parameter["default"],
                    required=False,
                    type=value_parse,
                    help="set %s, default is %s" % (
                        parameter["name"], parameter["default"]
                    )
                )
            else:
                subparser.add_argument(
                    parameter["name"],
                    type=value_parse,
                    help="set %s" % parameter["name"]
                )

    @staticmethod
    def func_spec(f):
        """获取函数参数表"""
        argspec = inspect.getfullargspec(f)
        argspec_args = argspec.args[1:] if is_classmethod(f) else argspec.args

        return f.__name__, args_zips(
            argspec_args, argspec.defaults
        ) + args_zips(
            argspec.kwonlyargs,
            argspec.kwonlydefaults.values() if argspec.kwonlydefaults else []
        )


class ParserGroup(argparse.ArgumentParser):
    """
    >>> class TestC(Configuration):
    ...     a = 1
    ...     b = 2
    >>> def test_f1(k=1):
    ...     return k
    >>> def test_f2(h=1):
    ...      return h
    >>> class TestC2(Configuration):
    ...     c = 3
    >>> parser1 = ConfigurationParser(TestC, commands=[test_f1])
    >>> parser2 = ConfigurationParser(TestC, commands=[test_f2])
    >>> pg = ParserGroup({"model1": parser1, "model2": parser2})
    >>> pg(["model1", "test_f1"])
    {'a': 1, 'b': 2, 'k': 1, 'subcommand': 'test_f1'}
    >>> pg("model2 test_f2")
    {'a': 1, 'b': 2, 'h': 1, 'subcommand': 'test_f2'}
    """

    def __init__(self, parsers: dict, prog=None,
                 usage=None,
                 description=None,
                 epilog=None, add_help=True):
        super(ParserGroup, self).__init__(
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            add_help=add_help
        )
        self.parsers = parsers

    def parse_args(self, args=None):
        args = sys.argv[1:] if args is None else args
        if args[0] in {"-h", "--help"}:  # pragma: no cover
            return super(ParserGroup, self).parse_args(args)
        else:
            target_parser = self.parsers[args[0]]
            real_args = args[1:]
            return target_parser.parse_args(real_args)

    def __call__(self, args: (str, list, None) = None) -> (dict, None):
        if isinstance(args, str):
            args = args.split(" ")
        args = sys.argv[1:] if args is None else args
        if args[0] in {"-h", "--help"}:  # pragma: no cover
            self.parse_args(args)
            return
        kwargs = self.parse_args(args)
        target_parser = args[0]
        return self.parsers[target_parser].parse(kwargs)

    def format_help(self) -> str:
        # hacking the original method
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)

        # description
        formatter.add_text(self.description)

        # positionals, optionals and user-defined groups
        for title, parser in self.parsers.items():
            formatter.start_section(title)
            formatter.add_text(parser.description)
            for action_group in parser._action_groups:
                formatter.start_section(action_group.title)
                formatter.add_text(action_group.description)
                formatter.add_arguments(action_group._group_actions)
                formatter.end_section()
            formatter.end_section()

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()


class Formatter(object):
    """
    以特定格式格式化字符串

    Examples
    --------
    >>> formatter = Formatter()
    >>> formatter("hello world")
    'hello world'
    >>> formatter = Formatter("hello {}")
    >>> formatter("world")
    'hello world'
    >>> formatter = Formatter("hello {} v{:.2f}")
    >>> formatter("world", 0.2)
    'hello world v0.20'
    >>> formatter = Formatter("hello {1} v{0:.2f}")
    >>> formatter(0.2, "world")
    'hello world v0.20'
    >>> Formatter.format(0.2, "world", formatter="hello {1} v{0:.3f}")
    'hello world v0.200'
    """

    def __init__(self, formatter: (str, None) = None):
        self.formatter = formatter

    def __call__(self, *format_string):
        if self.formatter is None:
            assert len(format_string) == 1, \
                "formatter is None, the input should be a single value, now is %s, " \
                "which has %s value" % (format_string, len(format_string))
            return format_string[0]
        else:
            return self.formatter.format(*format_string)

    @staticmethod
    def format(*format_string, formatter=None):
        formatter = Formatter(formatter)
        return formatter(*format_string)



# coding: utf-8
# 2020/8/16 @ tongshiwei
import logging
import os

import mxnet as mx
from mxnet.initializer import Initializer, Xavier, Uniform, Normal


def net_initialize(
        net, model_ctx,
        initializer: (str, Initializer, dict, list) = mx.init.Xavier(),
        select=None, logger=logging
):
    """
    初始化网络参数

    Notice
    ------
    The developer who modify this document should simultaneously modify the related function in glue

    """
    if isinstance(initializer, str):
        initializer = {
            "xaiver": Xavier(),
            "uniform": Uniform(),
            "normal": Normal()
        }[initializer]
    elif isinstance(initializer, dict):
        for _select, _initializer in initializer.items():
            net_initialize(net, model_ctx=model_ctx, initializer=_initializer, select=_select, logger=logger)
        return
    elif isinstance(initializer, list):
        if select is not None:
            assert len(select) == len(initializer)
            for _select, _initializer in zip(select, initializer):
                net_initialize(net, model_ctx=model_ctx, initializer=_initializer, select=_select, logger=logger)
        else:
            for _select, _initializer in initializer:
                net_initialize(net, model_ctx=model_ctx, initializer=_initializer, select=_select, logger=logger)
        return
    elif initializer is None or isinstance(initializer, Initializer):
        pass
    else:
        raise TypeError("initializer should be either str or Initializer, now is", type(initializer))

    logger.info("initializer: %s, select: %s, ctx: %s" % (initializer, select, model_ctx))
    net.collect_params(select).initialize(initializer, ctx=model_ctx)


def load_net(filename, net, ctx=mx.cpu(), allow_missing=False,
             ignore_extra=False):
    """
    Load the existing net parameters

    Parameters
    ----------
    filename: str
        The model file
    net: HybridBlock
        The network which has been initialized or
        loaded from the existed model
    ctx: Context or list of Context
            Defaults to ``mx.cpu()``.
    allow_missing: bool
    ignore_extra: bool

    Returns
    -------
    The initialized net
    """
    # 根据文件名装载已有的网络参数
    if not os.path.isfile(filename):
        raise FileExistsError("%s does not exist" % filename)
    net.load_parameters(filename, ctx, allow_missing=allow_missing,
                        ignore_extra=ignore_extra)
    return net



# coding: utf-8
# 2020/10/28 @ tongshiwei

from longling import path_append


def _model_path(model_path=None,
                model_dir="./",
                model_prefix="",
                model_suffix="",
                model_name="model.lgb"
                ):
    if model_path is None:
        return path_append(model_dir, model_prefix + model_name + model_suffix, to_str=True)
    return model_path


def train(
        batch_size=5000,
        epoch_num=10,
        params=None,
        model_path=None,
        model_dir="./",
        model_prefix="",
        model_suffix="",
        model_name="model.lgb",
        *args,
        **kwargs
):
    model_path = _model_path(model_path, model_dir, model_prefix, model_suffix, model_name)



# coding: utf-8
# create by tongshiwei on 2019/4/7

from .dataset import *
from .formatter.EvalFMT import *
from .formatter.EvalFormatter import *
from .monitor.LossMonitor import *
from .monitor.ProgressMonitor import *



# coding: utf-8
# create by tongshiwei on 2020-11-17

from longling import path_append

__all__ = ["epoch_params_filename", "get_epoch_params_filepath"]


def epoch_params_filename(model_name: str, epoch: int):
    return "%s-%04d.parmas" % (model_name, epoch)


def get_epoch_params_filepath(model_name: str, epoch: int, model_dir: str = "./"):
    return path_append(model_dir, epoch_params_filename(model_name, epoch), to_str=True)
